/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009-2010, Uwicore Laboratory (www.uwicore.umh.es),
 *                          University Miguel Hernandez, EU FP7 iTETRIS project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Ramon Bauza <rbauza@umh.es>
 */

#include "building-map-model.h"
#include "ns3/object-vector.h"
#include "ns3/simulator.h"
#include "ns3/config.h"
#include "ns3/log.h"
#include "segment.h"
#include "ns3/itetris-mobility-model.h"
#include "ns3/log.h"
#include <string>
#include <fstream>
#include <vector>

NS_LOG_COMPONENT_DEFINE ("BuildingMapModel");

namespace ns3 {

/**
 * \ingroup network
 * \brief private implementation detail of the NodeList API.
 */
class SegmentListPriv : public Object
{
public:
  static TypeId GetTypeId (void);
  SegmentListPriv ();
  ~SegmentListPriv ();
  
  uint32_t Add (Ptr<Segment> segment);
  BuildingMapModel::Iterator Begin (void) const;
  BuildingMapModel::Iterator End (void) const;
  Ptr<Segment> GetSegment (uint32_t n);
  uint32_t GetNSegments (void);
  std::vector<Ptr<Segment> > GetSegments ();
  bool GetVisibility (Vector posA, Vector posB);
  bool DoSegmentsIntersect (Ptr<Segment> seg1, Ptr<Segment> seg2);
  
  static Ptr<SegmentListPriv> Get (void);
  
private:
  virtual void DoDispose (void);
  static Ptr<SegmentListPriv> *DoGet (void);
  static void Delete (void);
  std::vector<Ptr<Segment> > m_segments;
};


NS_OBJECT_ENSURE_REGISTERED (SegmentListPriv);

TypeId 
SegmentListPriv::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::SegmentListPriv")
    .SetParent<Object> ()
    .AddAttribute ("BuildingMapModel","All segments in map",
                   ObjectVectorValue (),
                   MakeObjectVectorAccessor (&SegmentListPriv::m_segments),
                   MakeObjectVectorChecker<Segment> ())
    ;
  return tid;
}

Ptr<SegmentListPriv> 
SegmentListPriv::Get (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  return *DoGet ();
}

Ptr<SegmentListPriv> *
SegmentListPriv::DoGet (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  static Ptr<SegmentListPriv> ptr = 0;
  if (ptr == 0)
    {
      ptr = CreateObject<SegmentListPriv> ();
      Config::RegisterRootNamespaceObject (ptr);
      Simulator::ScheduleDestroy (&SegmentListPriv::Delete);
    }
  return &ptr;
}

void
SegmentListPriv::Delete (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  Config::UnregisterRootNamespaceObject (Get ());
  (*DoGet ()) = 0;
}

SegmentListPriv::SegmentListPriv ()
{
  NS_LOG_FUNCTION (this);
}

SegmentListPriv::~SegmentListPriv ()
{
  NS_LOG_FUNCTION (this);
}

void
SegmentListPriv::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();
  for (std::vector<Ptr<Segment> >::iterator i = m_segments.begin ();
       i != m_segments.end (); i++)
    {
      Ptr<Segment> segment = *i;
      segment->Dispose ();
      *i = 0;
    }
  m_segments.erase (m_segments.begin (), m_segments.end ());
  Object::DoDispose ();
}

uint32_t
SegmentListPriv::Add (Ptr<Segment> segment)
{
  uint32_t index = m_segments.size ();
  m_segments.push_back (segment);
  Simulator::ScheduleWithContext (index, TimeStep (0), &Segment::Initialize, segment);
  return index;
}

BuildingMapModel::Iterator
SegmentListPriv::Begin (void) const
{
  return m_segments.begin ();
}

BuildingMapModel::Iterator
SegmentListPriv::End (void) const
{
  return m_segments.end ();
}

uint32_t
SegmentListPriv::GetNSegments (void)
{
  return m_segments.size ();
}

Ptr<Segment>
SegmentListPriv::GetSegment (uint32_t n)
{
  NS_ASSERT_MSG (n < m_segments.size (), "Segment index " << n <<
                 " is out of range (only have " << m_segments.size () << " segments).");
  return m_segments.at (n);
}

std::vector<Ptr<Segment> >
SegmentListPriv::GetSegments()
{
  return m_segments;
}

// Copied from the bool BuildingMapModel::GetVisibility function
bool
SegmentListPriv::GetVisibility (Vector posA, Vector posB)
{
  bool vis = true;
  
  // Do not change the following constructor!!!
  Ptr<Segment> seg = CreateObject <Segment> ();  
  seg->x1 = posA.x;
  seg->y1 = posA.y;
  seg->x2 = posB.x;
  seg->y2 = posB.y;
          
  std::vector<Ptr<Segment> > segms = GetSegments();

  for (std::vector<Ptr<Segment> >::iterator it = segms.begin();
          it != segms.end(); ++it)
  {
    Ptr<Segment> segm = *it;
    NS_LOG_DEBUG ( "Segment id: " << segm->GetId() << " edges: x1: " << segm->x1
            << " y1: " << segm->y1 << " x2: " << segm->x2 << " y2: " << segm->y2 << "\n");
    if ( DoSegmentsIntersect (seg, *it) )
      {
        vis = false;
        break;
      }
  }
  NS_LOG_DEBUG ( "Num of segments: " << GetNSegments());
  NS_LOG_DEBUG ("Visibility = " << vis);
  return vis;
}

bool 
SegmentListPriv::DoSegmentsIntersect (Ptr<Segment> seg1, Ptr<Segment> seg2)
{
  // http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline2d/

  bool inter = false;
  double denom = ((seg2->y2-seg2->y1)*(seg1->x2-seg1->x1))-((seg2->x2-seg2->x1)*(seg1->y2-seg1->y1));
  double numa = (seg2->x2-seg2->x1)*(seg1->y1-seg2->y1)-(seg2->y2-seg2->y1)*(seg1->x1-seg2->x1);
  double numb = (seg1->x2-seg1->x1)*(seg1->y1-seg2->y1)-(seg1->y2-seg1->y1)*(seg1->x1-seg2->x1);  
  
  if (denom != 0.0)
  {
    double ua = numa/denom;
    double ub = numb/denom;
    if ( ua <= 1 && ua >= 0 && ub <= 1 && ub >= 0 ) // Segments intersect each other
      {
        inter = true;
        NS_LOG_DEBUG ("Segments intersecting");
        NS_LOG_DEBUG ("Seg1=(" <<seg1->x1 << "," <<seg1->y1 << ") ("<<seg1->x2 << "," << seg1->y2 << ")");
        NS_LOG_DEBUG ("Seg2=(" <<seg2->x1 << "," <<seg2->y1 << ") ("<<seg2->x2 << "," << seg2->y2 << ")");
      }
  }
  else if ( (numa == 0.0) && (numb == 0.0) )
  {
    inter = true;
    // NS_FATAL_ERROR ("BuildingMapModel::DoSegmentsIntersect - Tx and Rx are in the same position as one of the buildings in the scenario");
    NS_LOG_DEBUG ("Segments are on top of each other!");
  }

  return inter;
}
}

namespace ns3 {

//NS_OBJECT_ENSURE_REGISTERED (BuildingMapModel);

TypeId 
BuildingMapModel::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::BuildingMapModel")
    .SetParent<VisibilityModel> ()
    .AddConstructor<BuildingMapModel> ()
    .AddAttribute ("RoadMap","Road Map pointer",
                    PointerValue(0),
                    MakePointerAccessor(&BuildingMapModel::SetGridMap),
                    MakePointerChecker<GridMap>())
    ;
  return tid;
}

void
BuildingMapModel::SetGridMap (Ptr<GridMap> map)
{
  m_map = map;
}

BuildingMapModel::BuildingMapModel ()
{}

BuildingMapModel::~BuildingMapModel ()
{}

BuildingMapModel::BuildingMapModel (std::string path, Ptr<GridMap> map)
{
  m_path = path;
  m_map = map;
}

BuildingMapModel::BuildingMapModel (std::string path)
{
  m_path = path;
}

uint32_t
BuildingMapModel::Add (Ptr<Segment> segment)
{
  return SegmentListPriv::Get ()->Add (segment);
}
BuildingMapModel::Iterator
BuildingMapModel::Begin (void)
{
  return SegmentListPriv::Get ()->Begin ();
}
BuildingMapModel::Iterator
BuildingMapModel::End (void)
{
  return SegmentListPriv::Get ()->End ();
}
Ptr<Segment>
BuildingMapModel::GetSegment (uint32_t n)
{
  return SegmentListPriv::Get ()->GetSegment (n);
}
uint32_t
BuildingMapModel::GetNSegments (void)
{
  return SegmentListPriv::Get ()->GetNSegments ();
}

std::vector<Ptr<Segment> >
BuildingMapModel::GetSegments (void)
{
  return SegmentListPriv::Get ()->GetSegments ();
}

void 
BuildingMapModel::InitializeVisibilityModel (void)
{
  std::ifstream* m_is = new std::ifstream ();
  m_is->open (m_path.c_str (), std::ios::in);
  m_is->seekg (0);
  
  double x1 = 0;
  double y1 = 0;
  double x2 = 0;
  double y2 = 0;

  while (m_is->good())
    {
      *m_is >> x1 >> y1 >> x2 >> y2;
      CreateObject <Segment> (x1, y1, x2, y2);   
    }
  NS_LOG_DEBUG ("Num of Segments = " << GetNSegments());
  m_is->close ();
  delete m_is;
  m_is = 0;
}

void
BuildingMapModel::GetNlosDistances(Ptr<MobilityModel> a, Ptr<MobilityModel> b, double &dist1, double &dist2) const
{
//  dist1 = dist2 = a->GetDistanceFrom (b) / sqrt(2);
  NS_LOG_DEBUG ("BuildingMapModel::GetNlosDistances");

	Vector posA = a->GetPosition();
    Vector posB = b->GetPosition();

    int posA_X = m_map->GetRoadXFromVehicle(posA.x);
    int posA_Y = m_map->GetRoadYFromVehicle(posA.y);
	NS_LOG_DEBUG ("Node A: "<< posA << " posA_X = " << posA_X << " posA_Y = " << posA_Y);

	int posB_X = m_map->GetRoadXFromVehicle(posB.x);
    int posB_Y = m_map->GetRoadYFromVehicle(posB.y);
	NS_LOG_DEBUG ("Node B: "<< posB << " posB_X = " << posB_X << " posB_Y = " << posB_Y);


//    dist1 = m_map->GetLosDistance();
//    dist2 = m_map->GetNLostDistance();

    m_map->GetNLOSDistance(posA, posA_X, posA_Y, posB, posB_X, posB_Y, dist1, dist2);

	NS_LOG_DEBUG ("Distances d1 = " << dist1 << " d2 = " << dist2);
}

bool
BuildingMapModel::GetVisibility (Ptr<MobilityModel> a, Ptr<MobilityModel> b) const
{
  bool vis = true;
  Vector posA = a->GetPosition ();
  Vector posB = b->GetPosition ();
  
  // Do not change the following constructor!!!
  Ptr<Segment> seg = CreateObject <Segment> ();  
  seg->x1 = posA.x;
  seg->y1 = posA.y;
  seg->x2 = posB.x;
  seg->y2 = posB.y;
          
  std::vector<Ptr<Segment> > segms = GetSegments();

  for (std::vector<Ptr<Segment> >::iterator it = segms.begin();
          it != segms.end(); ++it)
  {
    Ptr<Segment> segm = *it;
    NS_LOG_DEBUG ( "Segment id: " << segm->GetId() << " edges: x1: " << segm->x1
            << " y1: " << segm->y1 << " x2: " << segm->x2 << " y2: " << segm->y2 << "\n");
    if ( DoSegmentsIntersect (seg, *it) )
      {
        vis = false;
        break;
      }
  }
  NS_LOG_DEBUG ( "Num of segments: " << GetNSegments());
  NS_LOG_DEBUG ("Visibility = " << vis);
  return vis;
}

bool 
BuildingMapModel::DoSegmentsIntersect (Ptr<Segment> seg1, Ptr<Segment> seg2) const
{
  // http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline2d/

  bool inter = false;
  double denom = ((seg2->y2-seg2->y1)*(seg1->x2-seg1->x1))-((seg2->x2-seg2->x1)*(seg1->y2-seg1->y1));
  double numa = (seg2->x2-seg2->x1)*(seg1->y1-seg2->y1)-(seg2->y2-seg2->y1)*(seg1->x1-seg2->x1);
  double numb = (seg1->x2-seg1->x1)*(seg1->y1-seg2->y1)-(seg1->y2-seg1->y1)*(seg1->x1-seg2->x1);  
  
  if (denom != 0.0)
    {
      double ua = numa/denom;
      double ub = numb/denom;
      if ( ua <= 1 && ua >= 0 && ub <= 1 && ub >= 0 ) // Segments intersect each other
        {
          inter = true;
          NS_LOG_DEBUG ("Segments intersecting");
          NS_LOG_DEBUG ("Seg1=(" <<seg1->x1 << "," <<seg1->y1 << ") ("<<seg1->x2 << "," << seg1->y2 << ")");
          NS_LOG_DEBUG ("Seg2=(" <<seg2->x1 << "," <<seg2->y1 << ") ("<<seg2->x2 << "," << seg2->y2 << ")");
        }
    }
  else if ( (numa == 0.0) && (numb == 0.0) )
    {
    inter = true;
//      NS_FATAL_ERROR ("BuildingMapModel::DoSegmentsIntersect - Tx and Rx are in the same position as one of the buildings in the scenario");
    }

  return inter;
}





bool 
BuildingMapModel::inLineOfSight (Vector a, Vector b)
{
  return SegmentListPriv::Get ()->GetVisibility (a, b);
}

} // namespace ns3
